from abc import ABCMeta, abstractmethod, abstractproperty, abstractstaticmethod
from typing import Any, AsyncGenerator, Generator, List, Optional, Tuple, Union
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import BaseEvent

from graia.broadcast.typing import T_Dispatcher, T_Dispatcher_Callable

class IDispatcherInterface(metaclass=ABCMeta):
    broadcast: "Broadcast"

    execution_contexts: List["ExecutionContext"]
    parameter_contexts: List["ParameterContext"]
    alive_generator_dispatcher: List[List[
        Tuple[Union[Generator, AsyncGenerator], bool]
    ]]
    
    @abstractmethod
    def start_execution(self, event: BaseEvent, dispatchers: List[T_Dispatcher], use_inline_generator: bool = False):
        pass

    @abstractmethod
    async def exit_current_execution(self):
        pass
    
    @abstractmethod
    async def alive_dispatcher_killer(self):
        pass
    
    def inject_local_raw(self, *dispatchers: List[T_Dispatcher], source: Any = None):
        for dispatcher in dispatchers[::-1]:
            self.parameter_contexts[-1].dispatchers.insert(0, dispatcher)
        always_dispatchers = self.execution_contexts[-1].always_dispatchers

        for i in dispatchers:
            if getattr(i, "always", False):
                always_dispatchers.add(i)

    def inject_execution_raw(self, *dispatchers: List[T_Dispatcher]):
        for dispatcher in dispatchers:
            self.execution_contexts[-1].dispatchers.insert(0, dispatcher)
        always_dispatchers = self.execution_contexts[-1].always_dispatchers

        for i in dispatchers:
            if getattr(i, "always", False):
                always_dispatchers.add(i)

    def inject_global_raw(self, *dispatchers: List[T_Dispatcher]):
        #self.dispatchers.extend(dispatchers)
        for dispatcher in dispatchers[::-1]:
            self.execution_contexts[0].dispatchers.insert(1, dispatcher)
        always_dispatchers = self.execution_contexts[-1].always_dispatchers

        for i in dispatchers:
            if getattr(i, "always", False):
                always_dispatchers.add(i)
    
    @abstractproperty
    def dispatcher_sources(self) -> List["DispatcherSource"]:
        pass
    
    @abstractmethod
    def dispatcher_pure_generator(self) -> Generator[None, None, T_Dispatcher]:
        pass

    @abstractmethod
    def dispatcher_generator(self) -> Generator[None, None, Tuple[T_Dispatcher, T_Dispatcher_Callable, Any]]:
        pass
    
    @abstractproperty
    def name(self) -> str:
        pass
    
    @abstractproperty
    def annotation(self) -> Any:
        pass
    
    @abstractproperty
    def default(self) -> Any:
        pass
    
    @abstractproperty
    def _index(self) -> int:
        pass
    
    @abstractproperty
    def global_dispatcher(self) -> List[T_Dispatcher]:
        pass
    
    @abstractproperty
    def event(self) -> BaseEvent:
        pass
    
    @abstractmethod
    async def execute_dispatcher_callable(self, dispatcher_callable: T_Dispatcher_Callable) -> Any:
        pass

    @abstractmethod
    async def lookup_param(self, name: str, annotation: Any, default: Any) -> Any:
        pass
    
    @abstractmethod
    async def lookup_using_current(self) -> Any:
        pass

    @abstractmethod
    async def lookup_by(self, dispatcher: T_Dispatcher, name: str, annotation: Any, default: Any) -> Any:
        pass

    @abstractmethod
    async def lookup_by_directly(self, dispatcher: T_Dispatcher, name: str, annotation: Any, default: Any) -> Any:
        pass

    @abstractproperty
    def has_current_exec_context(self) -> bool:
        pass
    
    @abstractproperty
    def has_current_param_context(self) -> bool:
        pass
    
    @abstractstaticmethod
    def preprocess_param_annotation(self, annotation: Any) -> Any:
        pass

from graia.broadcast.entities.source import DispatcherSource
from graia.broadcast.entities.context import ExecutionContext, ParameterContext


from typing import List

from graia.broadcast.entities.mapping_rule import MappingRule

from ..entities.dispatcher import BaseDispatcher
from ..entities.signatures import Force
from ..abstract.interfaces.dispatcher import IDispatcherInterface

def SimpleMapping(rules: List[MappingRule]):
    class mapping_dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: IDispatcherInterface):
            for rule in rules:
                if rule.mode(interface):
                    return Force(rule.value)
    return mapping_dispatcher

def Hook(condition, fixer):
    class hook_dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: IDispatcherInterface):
            if condition(interface):
                return fixer(interface.lookup_param(
                    interface.name,
                    interface.annotation,
                    interface.default
                ))
    return hook_dispatcher


import weakref

from typing import Any, List, Set, Type, Union
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.source import DispatcherSource
from .event import BaseEvent

from ..typing import T_Dispatcher

class ExecutionContext:
  source: DispatcherSource[T_Dispatcher, "ExecutionContext"]
  always_dispatchers: Set[Union[BaseDispatcher, Type[BaseDispatcher]]]
  event: BaseEvent
  inline_generator: bool
  _index: int

  def __init__(self, dispatchers: List[T_Dispatcher], event: BaseEvent, inline_generator: bool = False) -> None:
    self.source = DispatcherSource(dispatchers, weakref.ref(self))
    self.event = event
    self.inline_generator = inline_generator
    self._index = 0
  
    self.always_dispatchers = set()
    for i in dispatchers:
      if isinstance(i, BaseDispatcher) and i.always:
        self.always_dispatchers.add(i)
  
  @property
  def dispatchers(self):
    return self.source.dispatchers

class ParameterContext:
  def __init__(self, name, annotation, default, dispatchers, optional=False) -> None:
    self.name, self.annotation, self.default, self.source = \
      name, annotation, default, DispatcherSource(dispatchers, weakref.ref(self))
    self.optional = optional
  
  def __repr__(self) -> str:
    return "<ParameterContext name={0} annotation={1} default={2} locald={3}"\
      .format(
        self.name, self.annotation, self.default, self.dispatchers)

  source: DispatcherSource[T_Dispatcher, "ParameterContext"]

  name: str
  annotation: Any
  default: Any

  optional: bool

  @property
  def dispatchers(self) -> List[T_Dispatcher]:
    return self.source.dispatchers





from typing import Any, Callable, Generic, Iterable, Iterator, List, TypeVar, Union
from graia.broadcast.abstract.interfaces.dispatcher import IDispatcherInterface
from graia.broadcast.typing import T_Dispatcher

T = TypeVar("T", bound=T_Dispatcher)
S = TypeVar("S")

class DispatcherSource(Generic[T, S], Iterable[T]):
    dispatchers: List[Union[T, Callable[[IDispatcherInterface], T]]]
    source: S

    def __init__(self, dispatchers: List[Union[T, Callable[[IDispatcherInterface], T]]], source: S = None) -> None:
        self.dispatchers = dispatchers
        self.source = source
    
    def __iter__(self) -> Iterator[T]:
        yield from self.dispatchers


from typing import Any, Callable, Type, Union
from graia.broadcast.entities.dispatcher import BaseDispatcher

T_Dispatcher = Union[
    Type[BaseDispatcher],
    BaseDispatcher,
    Callable[["DispatcherInterface"], Any]
]

T_Dispatcher_Callable = Callable[["DispatcherInterface"], Any]


import weakref
import itertools

from pydantic.errors import NoneIsAllowedError

from graia.broadcast.abstract.interfaces.dispatcher import IDispatcherInterface
from typing import Any, AsyncGenerator, Callable, Generator, Iterator, List, Optional, Tuple, Union, _GenericAlias
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import BaseEvent

from graia.broadcast.entities.context import ExecutionContext, ParameterContext
from graia.broadcast.entities.signatures import Force
from graia.broadcast.entities.source import DispatcherSource
from graia.broadcast.exceptions import OutOfMaxGenerater, RequirementCrashed
from graia.broadcast.typing import T_Dispatcher, T_Dispatcher_Callable
from graia.broadcast.utilles import is_asyncgener, isgeneratorfunction, run_always_await_safely

class EmptyEvent(BaseEvent):
    class Dispatcher:
        @staticmethod
        def catch(_):
            pass

class DispatcherInterface(IDispatcherInterface):
    @property
    def name(self) -> str:
        return self.parameter_contexts[-1].name
    
    @property
    def annotation(self) -> Any:
        return self.parameter_contexts[-1].annotation
    
    @property
    def default(self) -> Any:
        return self.parameter_contexts[-1].default
    
    @property
    def _index(self) -> int:
        return self.execution_contexts[-1]._index
    
    @property
    def event(self) -> BaseEvent:
        return self.execution_contexts[-1].event

    @property
    def global_dispatcher(self) -> List[T_Dispatcher]:
        return self.execution_contexts[0].dispatchers
    
    @property
    def has_current_exec_context(self) -> bool:
        return len(self.execution_contexts) >= 2
    
    @property
    def has_current_param_context(self) -> bool:
        return len(self.parameter_contexts) >= 2

    def __init__(self, broadcast_instance: "Broadcast") -> None:
        self.broadcast = broadcast_instance
        self.alive_generator_dispatcher = [[]]
        self.execution_contexts = [ExecutionContext([], EmptyEvent())]
        self.parameter_contexts = [ParameterContext(None, None, None, [])]

    async def __aenter__(self) -> "DispatcherInterface":
        return self
    
    async def __aexit__(self, _, exc: Exception, tb):
        await self.exit_current_execution()
        if tb is not None:
            raise exc.with_traceback(tb)

    def start_execution(self, event: BaseEvent, dispatchers: List[T_Dispatcher], use_inline_generator: bool = False) -> "DispatcherInterface":
        current_exec_context = ExecutionContext(dispatchers, event, use_inline_generator)
        # 这里本来应该要让所有 always 集中下的......因为性能实在是太过惨淡, 故放弃
        self.execution_contexts.append(current_exec_context)
        self.alive_generator_dispatcher.append([])
        return self
    
    async def exit_current_execution(self):
        if self.alive_generator_dispatcher[-1]:
          if self.execution_contexts[-1].inline_generator:
            if len(self.alive_generator_dispatcher) > 2: # 防止插入到保护区
              self.alive_generator_dispatcher[-2].extend(self.alive_generator_dispatcher[-1])
            else:
              raise ValueError("cannot cast to inline")
          else:
            await self.alive_dispatcher_killer()
        self.alive_generator_dispatcher.pop()

        for i in self.dispatcher_pure_generator():
            after_execute_call = getattr(i, "after_execute", None)
            if callable(after_execute_call):
                try:
                    await run_always_await_safely(after_execute_call)
                except:
                    pass
        
        self.execution_contexts.pop()
    
    async def alive_dispatcher_killer(self):
        for unbound_gen, is_async_gen in self.alive_generator_dispatcher[-1]:
            index = 0
            if is_async_gen:
                async for _ in unbound_gen:
                    if index == 15:
                        raise OutOfMaxGenerater("a dispatch as a sync generator had to stop: ", unbound_gen)
                    index += 1
            else:
                for _ in unbound_gen:
                    if index == 15:
                        raise OutOfMaxGenerater("a dispatch as a sync generator had to stop: ", unbound_gen)
                    index += 1
    
    @property
    def dispatcher_sources(self) -> List[DispatcherSource]:
        return [
            self.execution_contexts[0].source,
            self.parameter_contexts[-1].source,
            self.execution_contexts[-1].source
        ]
    
    def dispatcher_pure_generator(self) -> Generator[None, None, T_Dispatcher]:
        for source in self.dispatcher_sources:
            yield from source.dispatchers
    
    def dispatcher_generator(self, 
        source_from: Callable[["DispatcherInterface"], Iterator[DispatcherSource[T_Dispatcher, Any]]] = \
            lambda x: x.dispatcher_sources
    ) -> Generator[None, None, Tuple[T_Dispatcher, T_Dispatcher_Callable, DispatcherSource]]:
        always_dispatcher = self.execution_contexts[-1].always_dispatchers
        
        for source in source_from(self):
            for dispatcher in source.dispatchers:
                if dispatcher in always_dispatcher:
                    always_dispatcher.remove(dispatcher)
                
                dispatcher_callable: T_Dispatcher_Callable = None
                if dispatcher.__class__ is type and issubclass(dispatcher, BaseDispatcher):
                    dispatcher_callable = dispatcher().catch
                elif hasattr(dispatcher, "catch"):
                    dispatcher_callable = dispatcher.catch
                elif callable(dispatcher):
                    dispatcher_callable = dispatcher
                else:
                    raise ValueError("invaild dispatcher: ", dispatcher)
                
                yield (dispatcher, dispatcher_callable, source)

    async def execute_dispatcher_callable(self, dispatcher_callable: T_Dispatcher_Callable) -> Any:
        alive_dispatchers = self.alive_generator_dispatcher[-1]

        if is_asyncgener(dispatcher_callable):
            current_generator = dispatcher_callable(self).__aiter__()

            try:
                result = await current_generator.__anext__()
            except StopAsyncIteration:
                return
            else:
                alive_dispatchers.append((current_generator, True))
        elif isgeneratorfunction(dispatcher_callable):
            current_generator = dispatcher_callable(self).__iter__()

            try:
                result = current_generator.__next__()
            except StopIteration as e:
                result = e.value
            else:
                alive_dispatchers.append((current_generator, False))
        else:
            result = await run_always_await_safely(dispatcher_callable, self)
        
        return result

    @staticmethod
    def preprocess_param_annotation(annotation: Any) -> Tuple[Any, bool]:
        optional = False
        if isinstance(annotation, _GenericAlias):
            if annotation.__origin__ is Union and annotation.__args__[-1] is type(None):
                # 如果是 Optional, 则它的最后一位应为 NoneType.
                optional = True
                annotation = annotation.__args__[0]
            else:
                raise TypeError("cannot preprocess this annotation: {0}".format(annotation))

        return (annotation, optional)

    async def lookup_param(self,
        name: str, annotation: Any, default: Any,
        enable_extra_return: bool = False
    ) -> Union[Any, Tuple[Any, T_Dispatcher, DispatcherSource, List[T_Dispatcher]]]:
        annotation, is_optional_param = self.preprocess_param_annotation(annotation)
        self.parameter_contexts.append(ParameterContext(
            name, annotation, default, [], optional=is_optional_param
        ))

        result = None
        try:
            start_offset = self._index + int(bool(self._index))
            for self.execution_contexts[-1]._index, (dispatcher, dispatcher_callable, source) in \
                enumerate(itertools.islice(self.dispatcher_generator(),
                    start_offset, None, None
            ), start=start_offset):
                result = await self.execute_dispatcher_callable(dispatcher_callable)
                
                if result is None:
                    continue
                
                if result.__class__ is Force:
                    result = result.target
                
                self.execution_contexts[-1]._index = 0
                return (
                    result, weakref.ref(dispatcher), source,
                    list(map(lambda x: weakref.ref(x), itertools.islice(
                        self.dispatcher_generator(), start_offset, self._index
                    )))
                ) if enable_extra_return else result
            else:
                if is_optional_param:
                    self.execution_contexts[-1]._index = 0
                    return
                raise RequirementCrashed("the dispatching requirement crashed: ", self.name, self.annotation, self.default)
        finally:
            for dispatcher, dispatcher_callable, source in self.dispatcher_generator(
                lambda x: [DispatcherSource(self.execution_contexts[-1].always_dispatchers)]
            ):
                await self.execute_dispatcher_callable(dispatcher_callable)

            self.parameter_contexts.pop()
    
    async def lookup_using_current(self) -> Any:
        _, is_optional_param = self.preprocess_param_annotation(self.annotation)

        result = None
        try:
            for self.execution_contexts[-1]._index, (dispatcher, dispatcher_callable, source) in \
                enumerate(itertools.islice(self.dispatcher_generator(),
                    self._index + int(bool(self._index)), None, None
            ), start=self._index + int(bool(self._index))):
                result = await self.execute_dispatcher_callable(dispatcher_callable)
                
                if result is None:
                    continue
                
                if result.__class__ is Force:
                    result = result.target
                
                self.execution_contexts[-1]._index = 0
                return result
            else:
                if is_optional_param:
                    self.execution_contexts[-1]._index = 0
                    return
                raise RequirementCrashed("the dispatching requirement crashed: ", self.name, self.annotation, self.default)
        finally:
            for dispatcher, dispatcher_callable, source in self.dispatcher_generator(
                lambda x: [DispatcherSource(self.execution_contexts[-1].always_dispatchers)]
            ):
                await self.execute_dispatcher_callable(dispatcher_callable)

    async def lookup_by(self, dispatcher: T_Dispatcher,
        name: str, annotation: Any, default: Any,
        enable_extra_return: bool = False
    ) -> Union[Any, Tuple[Any, T_Dispatcher, DispatcherSource, List[T_Dispatcher]]]:
        annotation, is_optional_param = self.preprocess_param_annotation(annotation)
        self.parameter_contexts.append(ParameterContext(
            name, annotation, default, [], optional=is_optional_param
        ))

        result = None
        try:
            start_offset = self._index + int(bool(self._index))
            for self.execution_contexts[-1]._index, (dispatcher, dispatcher_callable, source) in \
                enumerate(itertools.islice(self.dispatcher_generator(lambda x: [DispatcherSource([dispatcher])]),
                    start_offset, None, None
            ), start=start_offset):
                result = await self.execute_dispatcher_callable(dispatcher_callable)
                
                if result is None:
                    continue
                
                if result.__class__ is Force:
                    result = result.target
                
                self.execution_contexts[-1]._index = 0
                return (
                    result, weakref.ref(dispatcher), source,
                    list(map(lambda x: weakref.ref(x), itertools.islice(
                        self.dispatcher_generator(), start_offset, self._index
                    )))
                ) if enable_extra_return else result
            else:
                if is_optional_param:
                    self.execution_contexts[-1]._index = 0
                    return
                raise RequirementCrashed("the dispatching requirement crashed: ", self.name, self.annotation, self.default)
        finally:
            for dispatcher, dispatcher_callable, source in self.dispatcher_generator(
                lambda x: [DispatcherSource(self.execution_contexts[-1].always_dispatchers)]
            ):
                await self.execute_dispatcher_callable(dispatcher_callable)

            self.parameter_contexts.pop()

    async def lookup_by_directly(self, dispatcher: T_Dispatcher, name: str, annotation: Any, default: Any) -> Any:
        annotation, is_optional_param = self.preprocess_param_annotation(annotation)
        self.parameter_contexts.append(ParameterContext(
            name, annotation, default, [], optional=is_optional_param
        ))

        if dispatcher.__class__ is type and issubclass(dispatcher, BaseDispatcher):
            dispatcher_callable = dispatcher().catch
        elif hasattr(dispatcher, "catch"):
            dispatcher_callable = dispatcher.catch
        elif callable(dispatcher):
            dispatcher_callable = dispatcher
        else:
            raise ValueError("invaild dispatcher: ", dispatcher)
    
        try:
            result = await self.execute_dispatcher_callable(dispatcher_callable)
            if result.__class__ is Force:
                return result.target

            return result
        finally:
            self.parameter_contexts.pop()


from typing import Any, Callable, List
from pydantic import BaseModel
from graia.broadcast.abstract.interfaces.dispatcher import IDispatcherInterface

class MappingRule(BaseModel):
    mode: Callable[[IDispatcherInterface], bool]
    value: Any

    @classmethod
    def nameEquals(cls, name, value):
        return cls(mode=lambda x: x.name == name, value=value)

    @classmethod
    def annotationEquals(cls, annotation, value):
        return cls(mode=lambda x: x.annotation == annotation, value=value)

    @classmethod
    def defaultEquals(cls, default, value):
        return cls(mode=lambda x: x.default == default, value=value)


import asyncio
import inspect
import traceback
from typing import (Callable, Dict, Generator, Hashable, List, NoReturn,
                    Optional, Tuple, Type, Union)

from iterwrapper import IterWrapper as iw

from graia.broadcast.entities.exectarget import ExecTarget

from .abstract.interfaces.dispatcher import IDispatcherInterface
from .interfaces.dispatcher import DispatcherInterface
from .entities.decorater import Decorater
from .entities.dispatcher import BaseDispatcher
from .entities.event import BaseEvent
from .entities.inject_rule import BaseRule
from .entities.listener import Listener
from .entities.namespace import Namespace
from .entities.signatures import Force, RemoveMe
from .exceptions import (DisabledNamespace, ExecutionStop, ExistedNamespace,
                         InvaildEventName, PropagationCancelled,
                         RegisteredEventListener, RequirementCrashed,
                         UnexistedNamespace)
from .interfaces.decorater import DecoraterInterface
from .utilles import (argument_signature, dispatcher_mixin_handler, group_dict,
                      isasyncgen, isgenerator, run_always_await_safely)
from .typing import T_Dispatcher
from .zone import Zone


class Broadcast:
  loop: asyncio.AbstractEventLoop

  default_namespace: Namespace
  namespaces: List[Namespace]
  listeners: List[Listener]

  # 规则驱动注入
  dispatcher_inject_rules: List[BaseRule] = []
  
  dispatcher_interface: DispatcherInterface

  debug_flag: bool

  use_dispatcher_statistics: bool
  use_reference_optimization: bool

  def __init__(self, *, loop: asyncio.AbstractEventLoop = None, debug_flag: bool = False,
    inject_rules: Optional[List[BaseRule]] = None,
    use_dispatcher_statistics: bool = False,
    use_reference_optimization: bool = False
  ):
    self.loop = loop or asyncio.get_event_loop()
    self.default_namespace = Namespace(name="default", default=True)
    self.debug_flag = debug_flag
    self.namespaces = []
    self.listeners = []
    self.dispatcher_inject_rules = inject_rules or []
    self.dispatcher_interface = DispatcherInterface(self)

    self.use_dispatcher_statistics = use_dispatcher_statistics

    if not use_dispatcher_statistics and use_reference_optimization:
      raise ValueError("the feature of reference optimization requires dispatcher statictics.")
    self.use_reference_optimization = use_reference_optimization

    @self.dispatcher_interface.inject_global_raw
    def _(interface: DispatcherInterface):
      if interface.annotation is interface.event.__class__:
        return interface.event
  
  def default_listener_generator(self, event_class) -> Listener:
    yield from (iw(self.listeners)
      .filter(lambda x: not x.namespace.hide) # filter for hide
      .filter(lambda x: not x.namespace.disabled) # filter for disabled
      .filter(lambda x: event_class in x.listening_events)
      .collect(list) # collect to a whole list
    )

  def addInjectionRule(self, rule: BaseRule) -> NoReturn:
    if rule in self.dispatcher_inject_rules:
      raise ValueError("this rule has already been added!")
    self.dispatcher_inject_rules.append(rule)
  
  def removeInjectionRule(self, rule: BaseRule) -> NoReturn:
    self.dispatcher_inject_rules.remove(rule)

  async def layered_scheduler(self,
    listener_generator: Generator[Listener, None, None],
    event: BaseEvent
  ):
    grouped: Dict[int, List[Listener]] = group_dict(listener_generator, lambda x: x.priority)
    for current_priority in sorted(grouped.keys()):
      current_group = grouped[current_priority]
      try:
        await asyncio.gather(*[
          self.Executor(
            target=i,
            event=event
          ) for i in current_group
        ])
      except Exception as e:
        if isinstance(e, PropagationCancelled):
          break

  async def Executor(self,
    target: Union[Callable, ExecTarget],
    event: BaseEvent,
    dispatchers: List[Union[
      Type[BaseDispatcher],
      Callable,
      BaseDispatcher,
    ]] = None,
    
    post_exception_event: bool = False,
    print_exception: bool = True,
    enableInternalAccess: bool = False,
    use_inline_generator: bool = False,

    use_dispatcher_statistics: bool = False,
    use_reference_optimization: bool = False
  ):
    from .builtin.event import ExceptionThrowed

    is_exectarget = isinstance(target, ExecTarget)
    is_listener = isinstance(target, Listener)

    use_dispatcher_statistics = self.use_dispatcher_statistics or use_dispatcher_statistics
    use_reference_optimization = self.use_reference_optimization or use_reference_optimization

    if is_listener:
      if target.namespace.disabled:
        raise DisabledNamespace("catched a disabled namespace: {0}".format(target.namespace.name))
    
    if use_dispatcher_statistics and not is_exectarget:
      raise TypeError("the feature of dispatcher statistics requires a ExecTarget object as the target.")
    if use_reference_optimization and not use_dispatcher_statistics:
      raise ValueError("the feature of reference optimization requires dispatcher statictics.")

    # 先生成 dispatchers
    _dispatchers = dispatcher_mixin_handler(event.Dispatcher)
    # 开始暴力注入
    if is_exectarget:
      if target.inline_dispatchers:
        _dispatchers = target.inline_dispatchers + _dispatchers
      if is_listener:
        if target.namespace.injected_dispatchers:
          _dispatchers = target.namespace.injected_dispatchers + _dispatchers
    if dispatchers:
      _dispatchers = dispatchers + _dispatchers

    target_callable = target.callable if is_exectarget else target
    parameter_compile_result = {}

    async with self.dispatcher_interface.start_execution(event, _dispatchers, use_inline_generator) as dii:
      try:
        catched_first_dispatcher = next(dii.dispatcher_pure_generator())
      except StopIteration:
        dei = DecoraterInterface(dii)  # pylint: disable=unused-variable
        self.dispatcher_interface.execution_contexts[0].source.dispatchers.insert(0, dei)
      else:
        if not isinstance(catched_first_dispatcher, DecoraterInterface):
          dei = DecoraterInterface(dii)  # pylint: disable=unused-variable
          self.dispatcher_interface.execution_contexts[0].source.dispatchers.insert(0, dei)
      # Decorater 的 Dispatcher 已经注入, 没他事了

      if enableInternalAccess or \
        (is_exectarget and \
          target.enable_internal_access):
        internal_access_mapping = {
          Broadcast: self,
          DispatcherInterface: dii,
          **({ # 当 target 为 Listener 时的特有解析
            target.__class__: target,
          } if is_exectarget else {}),
          **({
            Namespace: target.namespace,
            
          } if is_listener else {})
        }
        @dii.inject_execution_raw
        def _(interface: DispatcherInterface):
          return internal_access_mapping.get(interface.annotation)

      for injection_rule in self.dispatcher_inject_rules:
        if injection_rule.check(event, dii):
          dii.execution_contexts[-1].dispatchers.insert(1, injection_rule.target_dispatcher)

      if is_exectarget:
        whole_statistics: Dict[str, Dict[T_Dispatcher, Tuple[int, int]]] = target.dispatcher_statistics['statistics']
      else:
        whole_statistics = {}
      try:
        for name, annotation, default in argument_signature(target_callable):
          statistics: Dict[T_Dispatcher, Tuple[int, int]] = whole_statistics.setdefault(name, {})
          if use_dispatcher_statistics:
            if use_reference_optimization: # 启用被动性质的优化特性 引用缓存(Reference Cache)
              for dispatcher, this_statistics in sorted(statistics.items(), key=lambda x: x[1][0]/x[1][1]):
                this_statistics[1] += 1
                if dispatcher() is None: # ref 在对象 dead 时返回 None.
                  del statistics[dispatcher]
                  continue

                try:
                  result = await dii.lookup_by_directly(dispatcher(), name, annotation, default)
                except RequirementCrashed: # 忽略单个 Dispatcher 执行时若无法满足要求时所引发的错误
                  pass
                except:
                  traceback.print_exc()
                  raise
                else:
                  this_statistics[0] += 1
                  parameter_compile_result[name] = result
                  # 本参数已经解析完毕, 但很显然, python 没有指定对象式的 break, so, break_flag, so ugly :(
              else:
                if parameter_compile_result.get(name):
                  continue
  
            parameter_compile_result[name], target_dispatcher, \
              target_source, past_dispatchers = \
                await dii.lookup_param(
                  name, annotation, default,
                  enable_extra_return=True
                )
            
            # TODO: 判定 target_source 来源的可靠性(毕竟...)
            if target_source.source() is not None:
              # 1.成功给出结果的次数, 2.总共被 call 的次数
              statistics.setdefault(target_dispatcher, [0, 0])
              this_statistics = statistics[target_dispatcher]
              this_statistics[0] += 1
              this_statistics[1] += 1
  
              if len(statistics) > 1:
                past_set = set(statistics.keys())
                for past_dispatcher in past_dispatchers:
                  if past_dispatcher in past_set:
                    statistics[past_dispatcher][1] += 1
          else:
            parameter_compile_result[name] =\
              await dii.lookup_param(name, annotation, default)

        if is_exectarget:
          if target.headless_decoraters:
            for hl_d in target.headless_decoraters:
              await dii.lookup_param(None, None, hl_d)
      except ExecutionStop:
        raise
      except RequirementCrashed:
        traceback.print_exc()
        raise
      except Exception as e:
        if print_exception:
          traceback.print_exc()
        if post_exception_event:
          self.postEvent(ExceptionThrowed(
            exception=e,
            event=event
          ))
        raise

      try:
        result = await run_always_await_safely(
          target_callable, **parameter_compile_result
        )
      except ExecutionStop:
        raise # 直接抛出.
      except PropagationCancelled:
        raise # 防止事件广播
      except Exception as e:
        if print_exception:
          traceback.print_exc()
        if post_exception_event: # 如果没有referrer, 则广播事件, 如果有则向上抛出(防止重复抛出事件)
          self.postEvent(ExceptionThrowed(
            exception=e,
            event=event
          ))
        raise
      
      gener = None
      if [inspect.isgenerator, isgenerator][isinstance(result, Hashable)](result):
        gener = result
        try: result = next(gener)
        except StopIteration as e: result = e.value
        else: dii.alive_generator_dispatcher[-1].append(
          (gener, False)
        )
      elif [inspect.isasyncgen, isasyncgen][isinstance(result, Hashable)](result):
        gener = result
        try: result = await gener.__anext__()
        except StopAsyncIteration: return # value = None
        else: dii.alive_generator_dispatcher[-1].append(
          (gener, True)
        )

      if result.__class__ is Force:
        return result.content

      if result.__class__ is RemoveMe:
        if isinstance(target, Listener):
          if target in self.listeners:
            self.listeners.pop(self.listeners.index(target))
      return result

  def postEvent(self, event: BaseEvent):
    self.loop.create_task(self.layered_scheduler(
      listener_generator=self.default_listener_generator(event.__class__),
      event=event
    ))

  @staticmethod
  def event_class_generator(target=BaseEvent):
    for i in target.__subclasses__():
      yield i
      if i.__subclasses__():
        yield from Broadcast.event_class_generator(i)

  @staticmethod
  def findEvent(name: str):
    for i in Broadcast.event_class_generator():
      if i.__name__ == name:
        return i

  def getDefaultNamespace(self):
    return self.default_namespace
  
  def createNamespace(self, name, *, priority: int = 0, hide: bool = False, disabled: bool = False):
    if self.containNamespace(name):
      raise ExistedNamespace(name, "has been created!")
    self.namespaces.append(Namespace(name=name, priority=priority, hide=hide, disabled=disabled))
    return self.namespaces[-1]
  
  def removeNamespace(self, name):
    if self.containNamespace(name):
      for index, i in enumerate(self.namespaces):
        if i.name == name:
          self.namespaces.pop(index)
          return
    else:
      raise UnexistedNamespace(name)

  def containNamespace(self, name):
    for i in self.namespaces:
      if i.name == name:
        return True
    return False
  
  def getNamespace(self, name):
    if self.containNamespace(name):
      for i in self.namespaces:
        if i.name == name:
          return i
    else:
      raise UnexistedNamespace(name)

  def hideNamespace(self, name):
    ns = self.getNamespace(name)
    ns.hide = True

  def unhideNamespace(self, name):
    ns = self.getNamespace(name)
    ns.hide = False

  def disableNamespace(self, name):
    ns = self.getNamespace(name)
    ns.disabled = True

  def enableNamespace(self, name):
    ns = self.getNamespace(name)
    ns.disabled = False

  def containListener(self, target):
    for i in self.listeners:
      if i.callable == target:
        return True
    return False

  def getListener(self, target):
    for i in self.listeners:
      if i.callable == target:
        return i

  def removeListener(self, target):
    self.listeners.remove(target)

  def includeZone(self, target_zone: Zone):
    if self.containNamespace(Zone.namespace.name):
      raise ValueError("the zone cannot import because its name has been used.")
    self.namespaces.append(target_zone.namespace)
    self.listeners.extend(target_zone.listeners)

  def receiver(self,
      event: Union[str, Type[BaseEvent]],
      priority: int = 16,
      dispatchers: List[Type[BaseDispatcher]] = [],
      namespace: Namespace = None,
      headless_decoraters: List[Decorater] = [],
      enable_internal_access: bool = False
  ):
    if isinstance(event, str):
      _name = event
      event = self.findEvent(event)
      if not event:
        raise InvaildEventName(_name, "is not vaild!")
    priority = (type(priority) == int) and priority or int(priority) # 类型转换
    def receiver_wrapper(callable_target):
      may_listener = self.getListener(callable_target)
      if not may_listener:
        self.listeners.append(Listener(
          callable=callable_target,
          namespace=namespace or self.getDefaultNamespace(),
          inline_dispatchers=dispatchers,
          priority=priority,
          listening_events=[event],
          headless_decoraters=headless_decoraters,
          enable_internal_access=enable_internal_access
        ))
      else:
        if event not in may_listener.listening_events:
          may_listener.listening_events.append(event)
        else:
          raise RegisteredEventListener(event.__name__, "has been registered!")
      return callable_target
    return receiver_wrapper



from typing import Callable, List, Type

from .dispatcher import BaseDispatcher
from .event import BaseEvent
from .namespace import Namespace
from .decorater import Decorater

from .exectarget import ExecTarget

class Listener(ExecTarget):
    def __init__(self,
        callable: Callable,
        namespace: Namespace,
        listening_events: List[Type[BaseEvent]],
        inline_dispatchers: List[BaseDispatcher] = None,
        headless_decoraters: List[Decorater] = None,
        priority: int = 16,
        enable_internal_access: bool = False,
    ) -> None:
        super().__init__(callable, inline_dispatchers, headless_decoraters, enable_internal_access)
        self.namespace = namespace
        self.listening_events = listening_events
        self.priority = priority

    namespace: Namespace
    listening_events: List[Type[BaseEvent]]
    priority: int = 16


from typing import Callable, Dict, List, Tuple, Union

from ..typing import T_Dispatcher
from .decorater import Decorater

class ExecTarget:
    def __init__(self,
        callable: Callable,
        inline_dispatchers: List[T_Dispatcher] = None,
        headless_decoraters: List[Decorater] = None,
        enable_internal_access: bool = False,
    ) -> None:
        self.callable = callable
        self.inline_dispatchers = inline_dispatchers or []
        self.headless_decoraters = headless_decoraters or []
        self.enable_internal_access = enable_internal_access

        self.dispatcher_statistics = {"total": 0, "statistics": {}}

    callable: Callable
    inline_dispatchers: List[T_Dispatcher] = []
    headless_decoraters: List[Decorater] = []
    enable_internal_access: bool = False

    dispatcher_statistics: Dict[str, Union[int, Dict]]